# -*- coding: utf-8 -*-
#
# Authors: Lukas Kopecky <l.kopecky22@imperial.ac.uk>
#          Timothy MD Ebbels 
#          Elizabeth J Want
#
# License: BSD 3-clause

import copy as deepcopy
import pandas as pd
import numpy as np
import statistics
from sklearn.metrics import (precision_score, recall_score, f1_score, roc_auc_score, accuracy_score, confusion_matrix,
                             ConfusionMatrixDisplay)
from sklearn.model_selection import KFold, GroupKFold, train_test_split
from sklearn.metrics import r2_score, root_mean_squared_error, mean_squared_error
from mbpls.mbpls import MBPLS
from sklearn.utils.validation import check_array, check_is_fitted
import matplotlib.pyplot as plt
from scipy import stats
from joblib import Parallel, delayed

class MamsiPls(MBPLS):
    """
    A class that extends the MB_PLS class by extra methods convenient in Chemometrics and Metabolomics research. 
    It is based on MB-PLS package: Baum et al., (2019). Multiblock PLS: Block dependent prediction modeling for Python.
    This wrapper has some extra methods convenient in Chemometrics and Metabolomics research.

    For a full list of methods, please refer to the MB-PLS class [documentation](https://mbpls.readthedocs.io/en/latest/index.html).

    Args:
        n_components (int, optional): A number of Latent Variables (LV). Defaults to 2.
        full_svd (bool, optional): Whether to use full singular value decomposition when performing the SVD method. 
            Set to False when using very large quadratic matrices (X). Defaults to False.
        method (str, optional): The method used to derive the model attributes. Options are 'UNIPALS', 'NIPALS', 'SIMPLS', 
            and 'KERNEL'. Defaults to 'NIPALS'.
        standardize (bool, optional): Whether to standardise the data (Unit-variance scaling). Defaults to True.
        max_tol (float, optional): Maximum tolerance allowed when using the iterative NIPALS algorithm. Defaults to 1e-14.
        nipals_convergence_norm (int, optional): Order of the norm that is used to calculate the difference of 
            the super-score vectors between subsequential iterations of the NIPALS algorithm. 
            Following orders are available:

            
            ord   | norm for matrices            | norm for vectors             |
            ----- | ---------------------------- | ---------------------------- |
            None  | Frobenius norm               |  2-norm                      |
            'fro' | Frobenius norm               | --                           |
            'nuc' | nuclear norm                 | --                           |
            inf   | max(sum(abs(x), axis=1))     | max(abs(x))                  |    
            -inf  | min(sum(abs(x), axis=1))     | min(abs(x))                  |
            0     | --                           | sum(x != 0)                  |
            1     | max(sum(abs(x), axis=0))     | as below                     |
            -1    | min(sum(abs(x), axis=0))     | as below                     |
            2     | 2-norm (largest sing. value) | as below                     |       
            -2    | smallest singular value      | as below                     |
            other | --                           | sum(abs(x)**ord)**(1./ord)   |

            Defaults to 2.
            
        calc_all (bool, optional): Whether to calculate all internal attributes for the used method. Some methods do not need
            to calculate all attributes (i.e., scores, weights) to obtain the regression coefficients used for prediction.
            Setting this parameter to False will omit these calculations for efficiency and speed. Defaults to True.
        sparse_data (bool, optional): NIPALS is the only algorithm that can handle sparse data using the method of H. Martens
            and Martens (2001) (p. 381). If this parameter is set to True, the method will be forced to NIPALS and sparse data
            is allowed. Without setting this parameter to True, sparse data will not be accepted. Defaults to False.
        copy (bool, optional): Whether the deflation should be done on a copy. Not using a copy might alter the input data
            and have unforeseeable consequences. Defaults to True.
     
    Attributes:
        n_components (int): Number of Latent Variables (LV).
        Ts_ (array): (X-Side) super scores [n,k]
        T_ (list): (X-Side) block scores [i][n,k]
        W_ (list): (X-Side) block weights [n][pi,k]
        A_ (array): (X-Side) block importances/super weights [i,k]
        A_corrected_ (array): (X-Side) normalized block importances (see mbpls documentation)
        P_ (list, block): (X-Side) loadings [i][pi,k]
        R_ (array): (X-Side) x_rotations R=W(PTW)-1
        explained_var_x_ (list): (X-Side) explained variance in X per LV [k]
        explained_var_xblocks_ (array): (X-Side) explained variance in each block Xi [i,k]
        beta_ (array): (X-Side) regression vector 𝛽 [p,q]
        U_ (array): (Y-Side) scoresInitialize [n,k]
        V_ (array): (Y-Side) loadings [q,k]
        explained_var_y_ (list): (Y-Side) explained variance in Y [k]
    """

    def __init__(self, n_components=2, full_svd=False, method='NIPALS', standardize=True, max_tol=1e-14,
                 nipals_convergence_norm=2, calc_all=True, sparse_data=False, copy=True):
        
        super().__init__(n_components, full_svd, method, standardize, max_tol, nipals_convergence_norm,
                         calc_all, sparse_data, copy)
        

    def estimate_lv(self, x, y, groups=None, max_components=10, classification=True, metric='auc', 
                    method='kfold', n_splits=5, repeats=100, test_size=0.2, random_state=42, 
                    plateau_threshold=0.01, increase_threshold=0.05, get_scores=False, savefig=False, n_jobs=-1,  **kwargs):
        """A method to estimate the number of latent variables (LVs)/components in the MB-PLS model.
           The method is based on cross-validation (k-fold or Monte Carlo) and combined with an outer loop with increasing number of LVs.
           LV on which the model stabilises corresponds with the optimal number of LVs.

        Args:
            x (array or list['array']): All blocks of predictors x1, x2, ..., xn. Rows are observations, columns are features/variables.
            y (array): A 1-dim array of reference values, either continuous or categorical variable.
            groups (array, optional): If provided, cv iterator variant with non-overlapping groups. 
                Group labels for the samples used while splitting the dataset into train/test set.
                Defaults to None.
            max_components (int, optional): Maximum number of components for whic LV estimate is calculated. 
                Defaults to 10.
            classification (bool, optional): Whether to perfrom calssification or regression. Defaults to True.
            metric (str, optional): Metric to use to estimate the number of LVs; available options: [`AUC`, `precision`, `recall`, `f1`] for 
                categorical outcome variables and ['q2'] for continuous outcome variable. 
                Defaults to 'auc'.
            method (str, optional): Corss-validation method. Available options ['kfold', 'montecarlo']. Defaults to 'kfold'.
            n_splits (int, optional): Number of splits for k-fold cross-validation. Defaults to 5.
            repeats (int, optional): Number of train-test split repeats from Monte Carlo. Defaults to 100.
            test_size (float, optional): Test size for Monte Carlo. Defaults to 0.2.
            random_state (int, optional): Generates a sequence of random splits to control MCCV. Defaults to 42.
            plateau_threshold (float, optional): Maximum increase for a sequence of LVs to be considered a plateau. 
                Must be non-negative. 
                Defaults to 0.01.
            increase_threshold (float, optional): Minimum increase to be considered a bend. Must be non-negative.. Defaults to 0.05.
            get_scores (bool, optional): Whether to retun measured mean scores. Defaults to False.
            savefig (bool, optional): Whether to save the plot as a figure. If True, argument `fname` has to be provided. Defaults to False.
            n_jobs(int, optional): Number of workers (CPU cores) for multiprocessing, -1 utilises all available cores on a system. 
                Defaults to -1.
            **kwargs: Additional keyword arguments to be passed to plt.savefig(), fname required to save .

        Raises:
            ValueError: Incorrect metric for categorical outcome. Allowed values are: 'auc', 'precision', 'recall', 'specificity', 'f1', 'accuracy'.
            ValueError: Incorrect metric for continuous outcome. Allowed values are: 'q2'.
            ValueError: Invalid method. Available options are ['kfold', 'montecarlo'].

        Returns:
            pandas.DataFrame: Measured mean scores for test and train splits for all components.
        """

        check_is_fitted(self, 'beta_')

        # Validation of data inputs
        _x = x.copy()
        if isinstance(_x, list) and not isinstance(_x[0], list):
            pass
        else:
            _x = [_x]
        _y = y.copy()
        _y = check_array(_y, ensure_2d=False)

        # Validation in parameter inputs
        if classification:
            allowed_metrics = ['auc', 'precision', 'recall', 'specificity', 'f1', 'accuracy']
            if metric not in allowed_metrics:
                raise ValueError(f"Invalid metric for categorical outcome. Allowed values are: "
                                 f"{', '.join(allowed_metrics)}")
        else:
            metric = 'q2'
            allowed_metrics = ['q2']
            if metric not in allowed_metrics:
                raise ValueError(f"Invalid metric continuous outcome. Allowed values are: {', '.join(allowed_metrics)}")
    
        # Scores placeholders
        training_means = pd.DataFrame()
        testing_means = pd.DataFrame()

        # Outer loop. Calculate scores for each latent variable.
        for i in range(1, max_components + 1):

            self.n_components = i
            if method == 'kfold':
                test_scores, train_scores = self.kfold_cv(x, y, groups=groups, classification=classification, return_train=True, n_splits=n_splits, n_jobs=n_jobs)
            elif method == 'montecarlo':
                test_scores, train_scores = self.montecarlo_cv(x, y, groups=groups, classification=classification, return_train=True, 
                                                                           test_size=test_size, repeats=repeats, random_state=random_state, n_jobs=n_jobs)
            else:
                raise ValueError("Invalid method. Available options are ['kfold', 'montecarlo']")

            test_scores = test_scores.dropna(axis=0)
            train_scores = train_scores.dropna(axis=0)

            if len(training_means) == 0:
                testing_means = pd.DataFrame(test_scores.mean()).T
                training_means = pd.DataFrame(train_scores.mean()).T
            else:
                testing_means = pd.concat([testing_means, pd.DataFrame(test_scores.mean()).T])
                training_means = pd.concat([training_means, pd.DataFrame(train_scores.mean()).T])

        testing_means.reset_index(drop=True, inplace=True)
        training_means.reset_index(drop=True, inplace=True)

        # concatenate training and testing scores into one dataframe, add prefix training_ to training scores
        perf_scores = pd.concat([training_means.add_prefix('training_'), testing_means], axis=1)
        perf_scores.insert(0, 'Number of Components', range(1, max_components + 1))

        #  for regression, delete RMSE from the dataframe and rename training_q2 to r2
        if not classification:
            perf_scores.drop('training_rmse', axis=1, inplace=True)
            perf_scores.drop('rmse', axis=1, inplace=True)
            perf_scores.rename(columns={'training_q2': 'r2'}, inplace=True)
            
        # Select desired metric
        if metric == 'q2':
            _data = testing_means['q2']
        if metric == 'auc':
            _data = testing_means['roc_auc']
        if metric == 'precision':
            _data = testing_means['precision']
        if metric == 'recall':
            _data = testing_means['recall']
        if metric == 'f1':
            _data = testing_means['f1']
        if metric == 'accuracy':
            _data = testing_means['accuracy']
        if metric == 'specificity':
            _data = testing_means['specificity']

        # Estimate number of LVs
        try:
            bend = np.min(np.where(np.diff(_data) / _data[0] < increase_threshold)[0]) + 1
        except ValueError:
            # Handle the case where np.min() fails due to an empty array
            bend = 1
        plateau_range_start, plateau_range_end = self._find_plateau(_data, range_threshold=plateau_threshold)

        # Percentage for printed statements below
        increase = increase_threshold * 100

        # Plot the results
        perf_scores.plot.line(x='Number of Components', marker='.', figsize=(8, 6), grid=False)
        plt.xlim(0, max_components + 1)
        plt.xticks(np.arange(1, max_components + 1, 1.0))
        try:
            plt.axvline(plateau_range_start, linestyle='--', color='r', label='Plateau edge')
            plt.axvline(bend, linestyle='dotted', color='b', label='Bend')
            print(metric + " reaches bent (increase of less than {0}".format(increase) +
                  " % of previous value or decrease) at component {0}".format(bend))
            print(metric + " reaches plateau at component {0}".format(plateau_range_start))
            self.n_components = plateau_range_start
            self.fit(x, y)
            print("Model re-fitted with n_components =", self.n_components)
        except TypeError:
            print(metric + " reaches bend (increase of less than {0}".format(increase) +
                  " % of previous value or decrease) at component {0}".format(bend))
            plt.axvline(bend, linestyle='--', color='b', label='Bend')
            print('No plateau detected, consider exploring more latent variables.')
            self.n_components = bend
            self.fit(x, y)
            print("Model re-fitted with n_components =", self.n_components)
        plt.ylabel('Score')
        plt.xlabel('Number of Latent Variables')
        if method == 'kfold':
            title = 'Latent Variable Estimation' + ' (k-Fold) '
        else:
            title = 'Latent Variable Estimation' + ' (Monte Carlo) ' 
        plt.title(title)
        plt.legend()

        if savefig:
            plt.savefig(**kwargs)

        if get_scores:
            return perf_scores    


    def evaluate_class_model(self, x, y):
        """
        Evaluate classification MB-PLS model using a **testing** dataset.

        Args:
            x (array or list[array]): All blocks of predictors x1, x2, ..., xn. Rows are observations, columns are features/variables.
            y (array): 1-dim or 2-dim array of reference values - categorical variable.

        Returns:
            array: Predicted y variable based on training set predictors.
        """

        # Check if PLS model is fitted
        check_is_fitted(self, 'beta_')

        # Validate inputs
        _x = x.copy()
        if isinstance(_x, list) and not isinstance(_x[0], list):
            pass
        else:
            _x = [x]
        _y = y.copy()
        _y = check_array(_y, ensure_2d=False)

        # Predict test data
        y_predicted = self.predict(_x)

        # Evaluation metrics
        cm2 = confusion_matrix(y, np.where(y_predicted > 0.5, 1, 0))
        disp2 = ConfusionMatrixDisplay(confusion_matrix=cm2)
        disp2.plot()
        print('Scores for model with n_components =', self.n_components)
        print('\nAccuracy', round(accuracy_score(_y, np.where(y_predicted > 0.5, 1, 0)), 3))
        print('Precision', round(precision_score(_y, np.where(y_predicted > 0.5, 1, 0)), 3))
        print('Recall', round(recall_score(_y, np.where(y_predicted > 0.5, 1, 0)), 3))
        tn, fp, fn, tp = confusion_matrix(_y, np.where(y_predicted > 0.5, 1, 0)).ravel()
        print('Specificity', round(tn/(tn+fp), 3))
        print('F1 Score', round(f1_score(_y, np.where(y_predicted > 0.5, 1, 0)), 3))
        print('AUC', round(roc_auc_score(_y, y_predicted), 3))
        return y_predicted
    
    def evaluate_regression_model(self, x, y):
        """
        Evaluate regression MB-PLS model using a **testing** dataset.

        Args:
            x (array or list[array]): All blocks of predictors x1, x2, ..., xn. Rows are observations, columns are features/variables.
            y (array): 1-dim or 2-dim array of reference values - continuous variable.

        Returns:
            array: Predicted y variable based on training set predictors.
        """

         # Check if PLS model is fitted
        check_is_fitted(self, 'beta_')

        # Validate inputs
        _x = x.copy()
        if isinstance(_x, list) and not isinstance(_x[0], list):
            pass
        else:
            _x = [x]
        _y = y.copy()
        _y = check_array(_y, ensure_2d=False)

        # Predict test data
        y_predicted = self.predict(_x)

        # Evaluation metrics
        rmse = root_mean_squared_error(_y, y_predicted)
        print(f'Root Mean Squared Error: {rmse}')
        q2 = r2_score(_y, y_predicted)
        print(f'Q-squared: {q2}')

        # Plotting Regresseion model evaluation
        plt.figure(dpi=600, figsize=(10, 6))
        plt.scatter(y_predicted, _y)
        plt.ylabel('Ground Truth')
        plt.xlabel('Predicted')
        plt.title('Regression Model Evaluation')

        return y_predicted
    
    def kfold_cv(self, x, y, groups=None, classification=True, return_train=False, n_splits=5, n_jobs=-1):
        """
        Perform k-fold cross-validation for MB-PLS model.

        Args:
            x (array or list[array]): All blocks of predictors x1, x2, ..., xn. Rows are observations, columns are features/variables.
            y (array): 1-dim or 2-dim array of reference values, either continuous or categorical variable.
            groups (array, optional): Group labels for the samples used while splitting the dataset into train/test set.
                If provided, group k-fold is performed. 
                Defaults to None.
            classification (bool, optional): Whether the outcome is a categorical variable. Defaults to True.
            return_train (bool, optional): Whether to return evaluation metrics for training set. Defaults to False.
            n_splits (int, optional): Number of splits for k-fold cross-validation. Defaults to 5.
            n_jobs (int, optional): Number of workers (CPU cores) for multiprocessing, -1 utilises all available cores on a system. 
                Defaults to -1.

        Returns:
            pandas.DataFrame: Evaluation metrics for each k-fold split.
                if return_train is True, returns evaluation metrics for training set as well.
        """

        check_is_fitted(self, 'beta_')

        # Validate inputs
        x_cp = x.copy()
        if isinstance(x_cp, list) and not isinstance(x_cp[0], list):
            pass
        else:
            x_cp = [x_cp]
        y_cp = y.copy()
        y_cp = check_array(y_cp, ensure_2d=False)

        # if groups are provided, group k-fold is performed, otherwise sk-learn k-fold is used
        if groups is None:
            kf = KFold(n_splits=n_splits)
        else:
            kf = GroupKFold(n_splits=n_splits)

        scores = pd.DataFrame()    

        def _kfold (_x, _y, _j, _train_indices, _test_indices, _classification, _return_train):
        
        # (j, (train_indices, test_indices) in enumerate(kf.split(_y, groups=groups))):
            # Unwrap data perform test-train split
            x_train = [None] * len(_x)
            x_test = [None] * len(_x)
            train_test_data = [None] * len(_x)
            y_train = _y[_train_indices]
            y_test = _y[_test_indices]

            for k in range(len(_x)):
                x_train[k] = pd.DataFrame(_x[k]).iloc[_train_indices]  # filter training data by index and save in a new list
                x_test[k] = pd.DataFrame(_x[k]).iloc[_test_indices]  # filter testing data by index and save in a new list
                train_test_data[k] = pd.DataFrame(_x[k]).iloc[_train_indices]

            # Fit model and predict
            x_train_copy = deepcopy.deepcopy(x_train)
            self.fit_transform(x_train_copy, y_train)# for each n_components fit new model
    
            # Predict outcome based on training folds
            y_predicted = self.predict(x_test)
            predictions = [y_predicted]
            truths = [y_test]

            # add training scores
            if _return_train:
                y_predicted_train = self.predict(x_train)
                predictions.append(y_predicted_train)
                truths.append(y_train)

            # Classification model evaluation
            if _classification:
                predictions_cl = [np.where(y_predicted > 0.5, 1, 0)]
                if _return_train:
                    predictions_cl.append(np.where(y_predicted_train > 0.5, 1, 0))

                # Calculate evaluation metrics for testing and training sets
                for prediction_cl, prediction, truth, j in zip(predictions_cl, predictions, truths, [0,1]):
                    # Evaluation metrics
                    try:
                        accuracy = accuracy_score(truth, prediction_cl)
                    except ValueError:
                        accuracy = np.nan
                    try:
                        precision = precision_score(truth, prediction_cl)
                    except ValueError:
                        precision = np.nan
                    try:
                        recall = recall_score(truth, prediction_cl, zero_division=np.nan)
                    except ValueError:
                        recall = np.nan
                    try:
                        f1 = f1_score(truth, prediction_cl)
                    except ValueError:
                        f1 = np.nan
                    try:
                        tn, fp, _, _ = confusion_matrix(truth, prediction_cl, labels=[0, 1]).ravel()
                        specificity_score = tn/(tn+fp) if (tn + fp) != 0 else np.nan
                    except ValueError:
                        specificity_score = np.nan
                    try:
                        roc_auc = roc_auc_score(truth, prediction)
                    except ValueError:
                        roc_auc = np.nan

                    row = np.array([[precision, recall, specificity_score, f1, roc_auc, accuracy]])

                    if j == 0:
                        # save MCCV scores
                        test_score_row = row.copy()
                    else:
                        train_score_row = row.copy()

            # Regression model evaluation    
            else:

                for prediction, truth, j in zip(predictions, truths, [0,1]):
                    # Evaluation metrics
                    rmse = root_mean_squared_error(truth, prediction)
                    q2 = r2_score(truth, prediction)

                    row = np.array([[rmse, q2]])


                    if j == 0:
                        # save MCCV scores
                        test_score_row = row.copy()
                    else:
                        train_score_row = row.copy()

            _scores = test_score_row
            if _return_train:
                _train_scores = train_score_row

            if _return_train:
                return _scores, _train_scores 

            else:
                return _scores, None

         # Run parellelised funciton 
        results = Parallel(n_jobs=n_jobs)(delayed(_kfold)
        (_x = x_cp, _y = y_cp, _j=i, _train_indices=train_indices, _test_indices=test_indices, _classification=classification, _return_train=return_train) 
        for i, (train_indices, test_indices) in enumerate(kf.split(y_cp, groups=groups)))
        
        # Get scores unwrapped 
        scores, train_scores = zip(*results)
        scores = np.stack(scores, axis=1)

        # Add lables to the the outcome labels
        if classification:
            scores = pd.DataFrame(scores[0], columns=['precision', 'recall', 'specificity', 'f1', 'roc_auc', 'accuracy' ]) 
        else:
            scores = pd.DataFrame(scores[0], columns=['rmse', 'q2'])

        if return_train:
            train_scores = np.stack(train_scores, axis=1)
            if classification:
                train_scores = pd.DataFrame(train_scores[0], columns=['precision', 'recall', 'specificity', 'f1', 'roc_auc', 'accuracy' ]) 
            else:
                train_scores = pd.DataFrame(train_scores[0], columns=['rmse', 'q2'])

        if return_train:
            return scores, train_scores
        else:    
            return scores


    def montecarlo_cv(self, x, y, groups=None, classification=True, return_train=False, test_size=0.2, repeats=10, random_state=42, n_jobs=-1):
        """
        Evaluate MB-PLS model using Monte Carlo Cross-Validation (MCCV).

        Args:
            x (array or list[array]): All blocks of predictors x1, x2, ..., xn. Rows are observations, columns are featuress.
            y (array): 1-dim or 2-dim array of reference values - categorical variable.
            groups (array, optional): Group labels for the samples used while splitting the dataset into train/test set.
                If provided, group-train-test split will be used instead of train-test split for random splits. 
                Defaults to None.
            classification (bool, optional): Whether the outcome is a categorical variable. Defaults to True.
            return_train (bool, optional): Whether to return evaluation metrics for training set. Defaults to False.
            test_size (float, optional): Proportion of the dataset to include in the test split. Defaults to 0.2.
            repeats (int, optional): Number of MCCV repeats. Defaults to 10.
            random_state (int, optional): Generates a sequence of random splits to control MCCV. Defaults to 42.
            n_jobs (int, optional): Number of workers (CPU cores) for multiprocessing, -1 utilises all available cores on a system. 
                Defaults to -1.

        Returns:
            pandas.DataFrame: Evaluation metrics for each MCCV repeat.
                if return_train is True, returns evaluation metrics for training set as well.
        """

        # Check if PLS model is fitted
        check_is_fitted(self, 'beta_')

        # Validate inputs
        x_cp = x.copy()
        if isinstance(x_cp, list) and not isinstance(x_cp[0], list):
            pass
        else:
            x_cp = [x_cp]
        y_cp = y.copy()
        y_cp = check_array(y_cp, ensure_2d=False)

        # Generate random sequence of seeds for MCCV
        rng = np.random.RandomState(random_state)
        
        # Generate n random numbers
        random_numbers = rng.randint(1, 4294967296, size=repeats)

        # Placeholder for MCCV scores
        scores = pd.DataFrame()

        def _mccv(_x, _y, i, _groups, _classification, _return_train, _test_size, _random_numbers):


            if _groups is None:
                train, test, y_train, y_test = train_test_split(pd.DataFrame(_x[0]), _y, test_size=_test_size, random_state=_random_numbers[i])
                if isinstance(_x[0], pd.DataFrame):
                    x_train = [df.loc[train.index] for df in _x]
                    x_test = [df.loc[test.index] for df in _x]
                else:
                    x_train = [df[train.index] for df in _x]
                    x_test = [df[train.index] for df in _x]
            else:
                x_train, x_test, y_train, y_test = self.group_train_test_split(_x, _y, groups=_groups, test_size=_test_size, random_state=_random_numbers[i])

            # Fit model and predict
            x_train_copy = deepcopy.deepcopy(x_train)
            self.fit_transform(x_train_copy, y_train)

            # Predict outcome based on training folds
            y_predicted = self.predict(x_test)
            predictions = [y_predicted]
            truths = [y_test]

            # add training scores
            if _return_train:
                y_predicted_train = self.predict(x_train)
                predictions.append(y_predicted_train)
                truths.append(y_train)

            # Classification model evaluation
            if _classification:
                predictions_cl = [np.where(y_predicted > 0.5, 1, 0)]
                if _return_train:
                    predictions_cl.append(np.where(y_predicted_train > 0.5, 1, 0))

                # Calculate evaluation metrics for testing and training sets
                for prediction_cl, prediction, truth, j in zip(predictions_cl, predictions, truths, [0,1]):
                    # Evaluation metrics
                    try:
                        accuracy = accuracy_score(truth, prediction_cl)
                    except ValueError:
                        accuracy = np.nan
                    try:
                        precision = precision_score(truth, prediction_cl)
                    except ValueError:
                        precision = np.nan
                    try:
                        recall = recall_score(truth, prediction_cl, zero_division=np.nan)
                    except ValueError:
                        recall = np.nan
                    try:
                        f1 = f1_score(truth, prediction_cl)
                    except ValueError:
                        f1 = np.nan
                    try:
                        tn, fp, _, _ = confusion_matrix(truth, prediction_cl, labels=[0, 1]).ravel()
                        specificity_score = tn/(tn+fp) if (tn + fp) != 0 else np.nan
                    except ValueError:
                        specificity_score = np.nan
                    try:
                        roc_auc = roc_auc_score(truth, prediction)
                    except ValueError:
                        roc_auc = np.nan

                    row = np.array([[precision, recall, specificity_score, f1, roc_auc, accuracy]])

                    if j == 0:
                        # save MCCV scores
                        test_score_row = row.copy()
                    else:
                        train_score_row = row.copy()

            # Regression model evaluation    
            else:

                for prediction, truth, j in zip(predictions, truths, [0,1]):
                    # Evaluation metrics
                    rmse = root_mean_squared_error(truth, prediction)
                    q2 = r2_score(truth, prediction)

                    row = np.array([[rmse, q2]])

                    if j == 0:
                        # save MCCV scores
                        test_score_row = row.copy()
                    else:
                        train_score_row = row.copy()
            
            _scores = test_score_row
            if _return_train:
                _train_scores = train_score_row

            if _return_train:
                return _scores, _train_scores 

            else:
                return _scores, None

        # Run parellelised funciton 
        results = Parallel(n_jobs=n_jobs)(delayed(_mccv)(_x = x_cp, _y=y_cp, i=j, _groups=groups, _classification=classification, _return_train=return_train, _test_size=test_size, _random_numbers=random_numbers) for j in range(repeats))
        
        # Get scores unwrapped 
        scores, train_scores = zip(*results)
        scores = np.stack(scores, axis=1)

        # Add lables to the the outcome labels
        if classification:
            scores = pd.DataFrame(scores[0], columns=['precision', 'recall', 'specificity', 'f1', 'roc_auc', 'accuracy' ]) 
        else:
            scores = pd.DataFrame(scores[0], columns=['rmse', 'q2'])

        if return_train:
            train_scores = np.stack(train_scores, axis=1)
            if classification:
                train_scores = pd.DataFrame(train_scores[0], columns=['precision', 'recall', 'specificity', 'f1', 'roc_auc', 'accuracy' ]) 
            else:
                train_scores = pd.DataFrame(train_scores[0], columns=['rmse', 'q2'])

        # retun data
        if return_train:
            return scores, train_scores
        else:    
            return scores
            

    def mb_vip(self, plot=True, get_scores=False, savefig=False, **kwargs):
        """
        Multi-block Variable Importance in Projection (MB-VIP) for multiblock PLS model.

        Adaptation of C. Wieder et al., (2024). PathIntegrate, doi: 10.1371/journal.pcbi.1011814.

        Args:
            plot (bool, optional): Whether to plot MB-VIP scores. Defaults to True.
            get_scores (bool, optional): Whether to return MB-VIP scores. Defaults to False.
            savefig (bool, optional): Whether to save the plot as a figure. If True, argument `fname` has to be provided. 
                Defaults to False.
            **kwargs: Additional keyword arguments to be passed to plt.savefig(), fname required to save .            

        Returns:
            array: MB-VIP scores.
        """

        # Check is model is fitted
        check_is_fitted(self, 'beta_')

        # stack the weights from all blocks
        weights = np.vstack(self.W_)
        # calculate product of sum of squares of super scores and y loadings
        sum_squares = np.sum(self.Ts_ ** 2, axis=0) * np.sum(self.V_ ** 2, axis=0)
        # p = number of variables - stack the loadings from all blocks
        p = np.vstack(self.P_).shape[0]
        # VIP is a weighted sum of squares of PLS weights
        vip_scores = np.sqrt(p * np.sum(sum_squares * (weights ** 2), axis=1) / np.sum(sum_squares))

        # Plot VIP scores
        if plot:
            plt.plot(vip_scores, color='limegreen', linewidth=0.8)
            plt.title('Multi-block variable importance in projection')
            plt.ylabel('MB-VIP score')
            plt.xlabel('Feature index')

            if savefig:
                plt.savefig(**kwargs)

        # Return all MB-VIP scores
        if get_scores:
            return vip_scores
        
            
    def block_importance(self, block_labels=None, normalised=True, plot=True, get_scores=False, savefig=False, **kwargs):
        '''
        Calculate the block importance for each block in the multiblock PLS model and plot the results.
        
        Args:
            block_labels (list, optional): List of block names. If block names are not provided or they do not match the number 
                of blocks in the model, the plot will display labels as 'Block 1', 'Block 2', ... 'Block n'. Defaults to None.
            normalised (bool, optional): Whether to use normalised block importance. For more information see model attribute 
                ['A_Corrected_'](). Defaults to True.
            plot (bool, optional): Whether to render plot block importance. Defaults to True.
            get_scores (bool, optional): Whether to return block importance scores. Defaults to False.
            savefig (bool, optional): Whether to save the plot as a figure. If True, argument `fname` has to be provided. 
                Defaults to False.
            **kwargs: Additional keyword arguments to be passed to plt.savefig(), fname required to save.
        
        Returns:
            array: Block importance scores.
        '''
        
        # Check if PLS model is fitted
        check_is_fitted(self, 'beta_')

        # get the block importances
        if normalised:
            block_importance = self.A_corrected_
        else:
            block_importance = self.A_
        block_importance_t = block_importance.T

        if plot:
            # Number of groups and number of bars in each group
            num_groups = len(block_importance)
            num_bars = len(block_importance[0])

            # Set up the figure and axis
            fig, ax = plt.subplots()

            # Set the width of the bars
            bar_width = 0.15  # Decreased bar width

            # Calculate the total width for each group of bars
            total_bar_width = bar_width * num_bars

            # Increase spacing between groups
            group_spacing = 0.5

            # Set the positions for the bars with offset to prevent overlap
            bar_positions = np.arange(num_groups) * (total_bar_width + group_spacing)

            # Plot each group of bars
            for i in range(num_bars):
                ax.bar(bar_positions + i * bar_width, block_importance_t[i], width=bar_width, label=f'LV {i + 1}')

            # Set labels and title
            ax.set_xlabel('Blocks')
            ax.set_ylabel('Explained variance')
            ax.set_title('Block Importance')
            ax.set_xticks(bar_positions + (total_bar_width / 2))

            # Set block names as x-tick labels
            if block_labels is None:
                # call blocks 'Block 1', 'Block 2' etc. if no names are provided
                block_labels = [f'Block {i + 1}' for i in range(num_groups)]
                ax.set_xticklabels(block_labels)

            elif len(block_labels) != num_groups:
                # call blocks 'Block 1', 'Block 2' etc. if length of block names does not match the number of blocks
                block_labels = [f'Block {i + 1}' for i in range(num_groups)]
                ax.set_xticklabels(block_labels)
                raise ValueError('Number of block names must match the number of blocks')
            
            else:
                # set block names as x-tick labels
                ax.set_xticklabels(block_labels)

            # Add legend
            ax.legend()

            # Show the plot
            plt.tight_layout()  # Adjust layout to prevent clipping of labels
        

            if savefig:
                plt.savefig(**kwargs)

            plt.show()

        if get_scores:
            return block_importance
        

    def mb_vip_permtest(self, x, y, n_permutations=1000, return_scores=False, n_jobs=-1):
        """
        Calculate empirical p-values for each feature by permuting the Y outcome variable `n_permutations` times and
        refitting the model. The p-values for each feature are calculated by counting the number of trials with
        MB-VIP greater than or equal to the observed test statistic, and dividing this by `n_permutations`.

        Args:
            x (array or list[array]): All blocks of predictors x1, x2, ..., xn. Rows are observations, columns are features/variables.
            y (array): 1-dim or 2-dim array of reference values, either continuous or categorical variable.
            n_permutations (int, optional): Number of permutation tests. Defaults to 1000.
            return_scores (bool, optional): Whether to return MB-VIP scores for each permuted null model. Defaults to False.
            n_jobs (int, optional): Number of workers (CPU cores) for multiprocessing, -1 utilises all available cores on a system. 
                Defaults to -1.

        Returns:
            array: Returns an array of p-values for each feature. If `return_scores` is True, then a matrix of MB-VIP scores
            for each permuted null model is returned as well.
        """


        # Check is model is fitted
        check_is_fitted(self, 'beta_')

        # Validation of data inputs
        _x = deepcopy.deepcopy(x)  # deepcopy to prevent data leakage
        if isinstance(_x, list) and not isinstance(_x[0], list):
            pass
        else:
            _x = [_x]
        _y = y.copy()
        _y = check_array(_y, ensure_2d=False)

        # MB-VIP of observed model
        _vip = self.mb_vip(plot=False, get_scores=True)
        vip_obs = _vip[:, np.newaxis]

        def _fit_permute(x, y):
            y_perm = np.random.permutation(y)
            self.fit(x, y_perm)
            return self.mb_vip(plot=False, get_scores=True)

        _vip_null = Parallel(n_jobs=n_jobs)(delayed(_fit_permute)(_x, _y) for _ in range(n_permutations))
        vip_null = np.stack(_vip_null, axis=1)

        # Calculate empirical p-values
        vip_greater = np.sum(vip_null >= vip_obs, axis=1)
        p_vals = vip_greater/n_permutations

        # Return p-vales and MB-PLS scores for null models
        if return_scores:
            return p_vals, vip_null
        else:
            return p_vals

    @staticmethod
    def calculate_ci(data, ci_level=0.90, dropna=True):
        """
        Static Method
        
        Calculates mean, margin of error, and confidence interval for each column.

        Args:
            data (pd.DataFrame): The input DataFrame.
            ci_level (float, optional): The confidence level (e.g., 0.90, 0.95). Defaults to 0.90.
            dropna (bool, optional): Whether to drop rows containing NaNs. Defaults to True.

        Returns:
            pd.DataFrame: A DataFrame containing the calculated statistics for each column.
                        If dropna = False, and a column has less than 2 valid values after
                        dropping NaNs specific to that column, all the result values for that
                        column will be np.nan.
        """

        if dropna:
            _data = data.dropna()  # Drop rows with NaNs for consistent comparison
        else:
            _data = data  # Use the data as-is (handle NaNs later if needed)

        results = {}
        for col in _data.columns:
            col_data = _data[col]
            if not dropna: # if user did not specify dropna, handle nans in individual columns
                col_data = col_data.dropna() # drop for each column separately
            
            if len(col_data) < 2: # handle cases where after dropping nans for a column, there is only one element. 
                print(f"Warning: Column {col} has less than 2 valid values after NaN removal. Cannot calculate CI.")
                results[col] = {"mean": np.nan, "margin_of_error": np.nan, "CI_lower": np.nan, "CI_upper": np.nan}
                continue

            mean = np.mean(col_data)
            sem = stats.sem(col_data)
            ci = stats.t.interval(ci_level, len(col_data) - 1, loc=mean, scale=sem)  # len(col_data) - 1: Degrees of freedom
            error = (ci[1] - ci[0]) / 2
            results[col] = {"mean": mean, "margin_of_error": error, "CI_lower": ci[0], "CI_upper": ci[1]}

        return pd.DataFrame(results)

    @staticmethod
    def group_train_test_split(x, y, gropus=None, test_size=0.2, random_state=42):
        """
        Static Method

        Split the data into train and test sets based on the groups. The groups are split into train and test sets
        based on the `test_size` parameter. The function returns the train and test sets for the predictors and the
        response variable.

        Args:
            x (array or list[array]): All blocks of predictors x1, x2, ..., xn. Rows are samples, columns are features.
            y (array): 1-dim or 2-dim array of reference values, either continuous or categorical variable.
            groups (array, optional): Group labels for the samples used while splitting the dataset into train/test set. Defaults to None.
            test_size (float, optional): Proportion of the dataset to include in the test split. Defaults to 0.2.
            random_state (int, optional): Controls the shuffling applied to the data before applying the split. Defaults to 42.

        Returns:
            tuple: x_train, x_test, y_train, y_test
        """
    
        # Ensure x is a list of data frames
        if not isinstance(x, list):
            x = [x]

        if groups is not None:
            groups = check_array(groups, ensure_2d=False, dtype=None)

        # Split the groups into train and test
        unique_groups = pd.unique(groups)
        train_groups, test_groups = train_test_split(unique_groups, test_size=test_size, random_state=random_state)

        # Create masks for train and test
        train_mask = np.isin(groups, train_groups)
        test_mask = np.isin(groups, test_groups)

        # Split the data frames in x
        x_train = [df.loc[train_mask] for df in x]
        x_test = [df.loc[test_mask] for df in x]

        # Split y
        y_train = y[train_mask]
        y_test = y[test_mask]

        return x_train, x_test, y_train, y_test
    

    @staticmethod
    def _find_plateau(scores, range_threshold=0.01, consecutive_elements=3):
        """
        Private Method, Static Method
        
        Function to assist in finding a plateau in a sequence of LVs.

        Args:
            scores (list[float]): List of scores.
            range_threshold (float, optional): Maximum increase for a sequence of LVs to be considered a plateau. Defaults to 0.01.
            consecutive_elements (int, optional): Number of elements that need to be in a plateau. Defaults to 3.

        Returns:
            tuple: Beginning and end indices of the plateau.
        """

        n = len(scores)
        for i in range(1, n - consecutive_elements + 1):
            plateau = True
            for j in range(consecutive_elements - 1):
                diff = abs(scores[i + j] - scores[i + j - 1])
                if diff > range_threshold:
                    plateau = False
                    break
            if plateau:
                return i, i + consecutive_elements - 1

        # If no plateau is found, return a tuple of None
        return None, None



# -*- coding: utf-8 -*-
#
# Author:   Lukas Kopecky <l.kopecky22@imperial.ac.uk>
#           Timothy MD Ebbels
#           Elizabeth J Want
#
# License: BSD 3-clause

import pandas as pd
import numpy as np
import re
import warnings
import matplotlib.pyplot as plt
import pkg_resources
from scipy.cluster.hierarchy import dendrogram, linkage, fcluster
from scipy.spatial.distance import squareform
import seaborn as sns
import networkx as nx
from pyvis.network import Network
from IPython.display import display
from IPython.display import IFrame
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import silhouette_score, silhouette_samples


class MamsiStructSearch:
    """
    A class for performing structural search on multi-modal MS data using.
    The class allows to search for structural signatures in LC-MS data based on their m/z and RT.
    These structural signatures include isotopologues and adduct patterns.


    Attributes:
        assay_links (list): List of data frames containing links for each assay.
        intensities (numpy.ndarray): Array of LC-MS intensity data.
        rt_win (int): Retention time tolerance window.
        ppm (int): Mass-to-charge ratio (m/z) tolerance in ppm.
        feature_metadata (pandas.DataFrame): Data frame containing feature metadata extracted from column names.
        structural_links (pandas.DataFrame): Data frame containing structural clusters.

    Args:
        rt_win (int, optional): Retention time tolerance window. Defaults to 5.
        ppm (int, optional): Mass-to-charge ratio (m/z) tolerance in ppm. Defaults to 15.
    """

    def __init__(self, rt_win=5, ppm=15):
        self.assay_links = None
        self.intensities = None
        self.rt_win = rt_win
        self.ppm = ppm
        self.feature_metadata = None
        self.structural_links = None


    def load_msi(self, df):
        """
        Imports MSI intensity data and extracts feature metadata from column names.

        Args:
            df (pandas.DataFrame): Data frame with MSI intensity data.
                - rows: samples
                - columns: features (m/z peaks)
                    Column names in the format: *(m/z)* For example: *149.111*
        """

        _df = df.copy()

        # Extract column names
        names = _df.columns
        deconstructed = pd.DataFrame()
        for name in names:
            assay = name.split('_')[0]
            mz = float(name.split('_')[1])
            line = pd.DataFrame({
                'Feature': name,
                'Assay': assay,
                'RT': 1,
                'm/z': mz
            }, index=[0]
            )
            deconstructed = pd.concat([deconstructed, line])
        deconstructed.reset_index(inplace=True, drop=True)

        # Split metadata into assays
        data = []
        unique_assays = deconstructed['Assay'].unique()
        for assay in unique_assays:
            data.append(deconstructed[deconstructed['Assay'] == assay])

        # Save data
        self.feature_metadata = deconstructed
        self.assay_links = data
        self.intensities = _df


    def load_lcms(self, df):
        """
        Imports LC-MS intensity data and extracts feature metadata from column names.

        Args:
            df (pandas.DataFrame): Data frame with LC-MS intensity data.
                - rows: samples
                - columns: features (LC-MS peaks). Column names in the format: **(AssayName)_(RTsec)_(m/z)m/z**.
                        For example: **HPOS_233.25_149.111m/z**
        """

        _df = df.copy()

        # Extract column names
        _df_columns = _df.columns
        names = _df_columns.to_list()  # Extract names of features

        # Extract metadata from column names
        deconstructed = pd.DataFrame()
        for name in names:
            assay = name.split('_')[0]  
            rt = float(name.split('_')[1])  
            mz = name.split('_')[2]  
            mz = float(mz.split('m')[0])  
            line = pd.DataFrame({
                'Feature': name,
                'Assay': assay,
                'RT': rt,
                'm/z': mz
            }, index=[0]
            )
            deconstructed = pd.concat([deconstructed, line])
        deconstructed.reset_index(inplace=True, drop=True)

        # Split metadata into assays
        data = []
        unique_assays = deconstructed['Assay'].unique()
        for assay in unique_assays:
            data.append(deconstructed[deconstructed['Assay'] == assay])

        # Save data
        self.feature_metadata = deconstructed
        self.assay_links = data
        self.intensities = _df

    def get_structural_clusters(self, adducts='all', annotate=True):
        """
        Searches structural signatures in LC-MS data based on their m/z and RT. These structural signatures include 
        isotopologues, adduct patterns and cross-assay links.

        Args:
            adducts (str, optional): Define what type of adducts to . 
                Possible values are:
                - 'all': All adducts combinations (based on Fiehn Lab adduct calculator).
                - 'most-common': Most common adducts for ESI (based on Waters adducts documentation).
                Defaults to 'all'.
            annotate (bool, optional): Annotate significant features based on National Phenome Centre RIO data.
                Only to be run if the data was analysed by the National Phenome Centre or analysis followed their
                conventions and protocols. For more information see https://doi.org/10.1021/acs.analchem.6b01481 
                or https://phenomecentre.org.
                Uses semi-targeted annotations for selected compounds.
                Defaults to True.

        Returns:
            pandas.DataFrame or list(pandas.DataFrame): DataFrame of significant features with structural clusters.
        """

        # Check if metadata have been loaded
        if self.feature_metadata is None:
            warnings.simplefilter('error', RuntimeWarning)
            warnings.warn("No data loaded. Use 'load_lcms()' to load data.",
                          RuntimeWarning)

        # Get isotopologue and adduct clusters
        self._get_isotopologue_groups()
        self._get_adduct_groups()
        self._get_unified_struct_groups()

        # Get annotation from ROI files (NPC)
        if annotate:
            self._get_annotation()

        data = self.assay_links

        # PROCESS ASSAYS
        data_both = []
        iso_offset = 0
        adduct_offset = 0
        cluster_offset = 0
        for frame in data:

            working_frame = frame

            data_both.append(working_frame)


            # Update offsets for isotopologue and adduct clusters
            working_frame['Isotopologue group'] = working_frame['Isotopologue group'].apply(lambda x: x + iso_offset)
            # print(working_frame['Adduct group'].max())
            working_frame['Adduct group'] = working_frame['Adduct group'].apply(lambda x: x + adduct_offset)
            working_frame['Structural cluster'] = working_frame['Structural cluster'].apply(lambda x: x + cluster_offset)
            if not np.isnan(working_frame['Isotopologue group'].max()):
                iso_offset = working_frame['Isotopologue group'].max()  # Update offset
            if not np.isnan(working_frame['Adduct group'].max()):
                adduct_offset = working_frame['Adduct group'].max()
            cluster_offset = working_frame['Structural cluster'].max()  # Update offset
            
        data_both = pd.DataFrame(np.vstack(data_both), columns=data_both[1].columns)
        self.structural_links = data_both
        # Get cross-assay links
        self._get_cross_assay_links()
        
        return  self.structural_links    

    def _get_isotopologue_groups(self):
        """
        Private method.
        Search for isotopologue signature in individual assay data frames.
        """

        for index, frame in enumerate(self.assay_links):
            # Create a copy of data frame
            frame = frame.copy()
            # Sort copied data frame and create new column
            frame.sort_values(by='m/z', inplace=True)
            frame.reset_index(inplace=True, drop=True)
            frame['Isotopologue group'] = [np.nan] * len(frame)
            frame['Isotopologue pattern'] = [np.nan] * len(frame)

            # Group ID for new cluster
            iso_group = 1
            m_plus = 0

            # Loop through data frame. Add neutron mass to current value
            for i in range(len(frame)):
                current_mz = frame.loc[i, 'm/z']
                current_mz += 1.003355
                current_rt = frame.loc[i, 'RT']
    
                
                # Loop through data frame to find expected MZ
                for j in range(len(frame)):
                    expected_mz = frame.loc[j, 'm/z']
                    expected_rt = frame.loc[j, 'RT']

                    # Get RT and PPM differences between the peaks
                    ppm_diff = self._mean_ppm_diff(expected_mz, current_mz)
                    rt_diff = abs(current_rt - expected_rt)

                    # Check if RT and PPM are within tolerance and add group ID
                    if ppm_diff <= self.ppm and rt_diff <= self.rt_win:
                        # If NaN use a new cluster ID
                        if np.isnan(frame.loc[i, 'Isotopologue group']):
                            frame.at[i, 'Isotopologue group'] = iso_group
                            frame.at[j, 'Isotopologue group'] = iso_group
                            iso_group += 1
                            m_plus = 0  # Reset M+ counter if new isotopologue group is found

                        # If NOT NaN use current cluster ID
                        if not np.isnan(frame.loc[i, 'Isotopologue group']):
                            frame.at[j, 'Isotopologue group'] = frame.loc[i, 'Isotopologue group']
                            # Assign M and M+ values
                            frame.at[i, 'Isotopologue pattern'] = m_plus
                            m_plus += 1
                            frame.at[j, 'Isotopologue pattern'] = m_plus
            
            # Update Current DataFrame
            self.assay_links[index] = frame

    def _get_adduct_groups(self, adducts='all'):
        """
        Private method.
        Search for adduct grouping signatures within significant features. 
        The method uses a ppm window to search for adducts. 

        Args:
            adducts (str, optional): _description_. Defaults to 'all'.
        """

        for index, frame in enumerate(self.assay_links):

            frame_ = frame.copy()

            # Detect isotopologues within one loop

            # Get neutral masses for all adducts
            frame__ = self._get_neutral_mass(features=frame_, adducts=adducts)

            # Search for adducts in current DataFrame
            data_clusters_frame = self._search_main_adduct(frame__)

            # Combine isotopologue and adduct clusters into one DataFrame
            frame_2 = frame__.iloc[:, :6]
            working_frame = frame_2.merge(data_clusters_frame.iloc[:, [0, 7, 2, 3, 4, 5]], on='Feature', how='left')

            # Merge overlapping adduct clusters

            non_unique_features = working_frame['Feature'][working_frame['Feature'].duplicated(keep=False)].unique()
            for item in non_unique_features:
                # For all non-unique features, find all clusters they belong too
                fr_ = working_frame[working_frame['Feature'] == item].loc[:, ['Feature', 'Adduct group', 'Adduct']]
                combined_adduct = '/'.join(fr_['Adduct'])
                fr_['Adduct'] = combined_adduct

                #
                working_frame.reset_index(inplace=True, drop=True)
                fr_.reset_index(inplace=True, drop=True)
                for i in range(len(fr_)):
                    working_frame.loc[working_frame['Feature'] ==
                                      fr_.loc[i, 'Feature'], 'Adduct'] = fr_.loc[i, 'Adduct']

                # unify all overlapping cluster by assigning the lowest cluster values to all clusters
                for i in range(len(fr_) - 1):
                    working_frame['Adduct group'] = working_frame['Adduct group'].replace({fr_.iloc[i + 1, 1]: fr_.iloc[0, 1]})

                working_frame = working_frame.drop_duplicates(subset='Feature')  # Delete non unique Features

            self.assay_links[index] = working_frame
            # now load data below in the main loop as nothing is returned

    def _get_neutral_mass(self, features, adducts='all'):
        """
        Private method.
        Calculate potential neutral masses for all m/z features.

        Args:
            features (pandas.DataFrame): DataFrame containing m/z features.
            adducts (str, optional): DataFrame with ion masses and names. Defaults to 'all'.

        Returns:
            pandas.DataFrame: DataFrame with m/z and hypothetical neutral masses for given adducts.
        """
        
        # Load all files with "all" adducts
        if adducts == 'all':
            stream_all_adducts_pos = pkg_resources.resource_stream(__name__, 'Data/Adducts/all_adducts_pos.csv')

            adducts_positive = pd.read_csv(stream_all_adducts_pos)
            
            stream_all_adducts_neg = pkg_resources.resource_stream(__name__, 'Data/Adducts/all_adducts_neg.csv')
            adducts_negative = pd.read_csv(stream_all_adducts_neg)

        # Load files with the "most common" adducts
        else:
            # Load external adduct files
            stream_common_adducts_pos = pkg_resources.resource_stream(__name__, 'Data/Adducts/common_adducts_pos.csv')
            adducts_positive = pd.read_csv(stream_common_adducts_pos)
            
            stream_common_adducts_neg = pkg_resources.resource_stream(__name__, 'Data/Adducts/common_adducts_neg.csv')
            adducts_negative = pd.read_csv(stream_common_adducts_neg)

        df = features.copy()  # Copy features data frame
        df.reset_index(inplace=True, drop=True)
        assay = df.loc[0, 'Assay']
        if re.search(r'pos', assay,  re.IGNORECASE):
            adducts = adducts_positive
        elif re.search(r'neg', assay,  re.IGNORECASE):
            adducts = adducts_negative
        else:
            raise Exception("Ionisation mode has not been recognised. Please ensure that ionisation mode is "
                            "either 'POS' for positive or 'NEG' for negative types of ionisation")

        adducts = adducts.copy()  # Copy adduct ion masses
        adducts.reset_index(inplace=True, drop=True)
        for j in range(adducts.shape[0]):  # Calculate hypothetical neutral masses for all given adduct ...
            column = []
            for i in range(len(df)):  # ... and repeat for all peaks
                column.append((df.loc[i, 'm/z'] - adducts.loc[j, 'Mass']) / adducts.loc[j, '1/Charge'])  # Calculate U
            name = adducts.loc[j, 'Ion name']  # Get the column name from adduct name in the adduct file
            df.insert(features.shape[1] + j, name, column)  # Append column of neutral masses for given adduct to the DF
        return df

    def _search_main_adduct(self, x):
        """
        Private method.
        Search for main adducts ([M+H]+ / [M-H]-) in the given input.

        Args:
            x (DataFrame): The input DataFrame containing the data to search.

        Returns:
            DataFrame: A DataFrame containing the matches found for the main adducts.
        """
        
        frame_ = x.copy()

        # Outer Loop - go through main adducts
        matches = pd.DataFrame()
        cluster_flag = 1
        for index, row in frame_.iterrows():
            group = self._find_adduct_matches(frame_, row, cluster_flag)
            if len(group) > 0:
                matches = pd.concat([matches, group])
                cluster_flag = cluster_flag + 1
            else:
                matches = pd.concat([matches, group])
        return matches

    def _find_adduct_matches(self, frame_, row_, clust_flag, main_adduct=True):
        """
        Private method.
        Method searches for adducts within a given ppm window.

        Args:
            frame_ (pandas.DataFrame): DataFrame containing m/z values.
            row_ (pandas.Series): Row of the DataFrame.
            clust_flag (bool): Cluster flag.
            main_adduct (bool): Flag for main adducts.

        Returns:
            pandas.DataFrame: DataFrame with adducts.
        """

        main_adduct_index = 6  # This is a default column, needs to be updated if default adduct changes

        frame = frame_.copy()
        group = pd.DataFrame(columns=['Feature', 'm/z', 'Expected neutral mass',
                                      'Observed neutral mass', 'Neutral mass |difference ppm|',
                                      'Adduct', 'RT', 'Adduct group'
                                      ])  # Add groups one by one and see what the result looks like

        frame = frame[~(frame['Feature'] == row_['Feature'])]  # Delete current row from frame
        rt_now = row_['RT']  # Filter frame within RT window
        frame = frame.loc[(frame['RT'] >= rt_now - self.rt_win) & (frame['RT'] <= rt_now + self.rt_win)]  # RT WINDOW
        if row_['Isotopologue group'] > 0:  # Filter out all suspected isotopologues identical to current row
            frame = frame[~(frame['Isotopologue group'] == row_['Isotopologue group'])]
        val_current = row_.iloc[main_adduct_index]

        # Loop through what remained from frame
        for i in range(frame.shape[0]):  # Search in rows
            for j in range(main_adduct_index, frame.shape[1]):  # Search in columns  MUST BE UPDATED FOR FLEXIBILITY OF COLUMNS!
                ppm_diff1 = self._mean_ppm_diff(frame.iloc[i, j], val_current)
                if ppm_diff1 <= self.ppm:
                    line = pd.DataFrame({
                        'Feature': frame.iloc[i, 0],
                        'm/z': frame.iloc[i, 3],
                        'Expected neutral mass': val_current,
                        'Observed neutral mass': frame.iloc[i, j],
                        'Neutral mass |difference ppm|': ppm_diff1,
                        'Adduct': frame.columns[j],
                        'RT': frame.iloc[i, 2],
                        'Adduct group': clust_flag
                    }, index=[0])

                    # To avoid concatenating empty data frame
                    if len(group) == 0:
                        group = line.copy()
                    else:    
                        group = pd.concat([group, line])

        # If there are matches, add the current row into the group
        if len(group) > 0:
            line = pd.DataFrame({
                'Feature': row_['Feature'],
                'm/z': row_['m/z'],
                'Expected neutral mass': val_current,
                'Observed neutral mass': val_current,
                'Neutral mass |difference ppm|': 0,
                'Adduct': frame.columns[6],  # This is a default column, needs to be updated if default adduct changes
                'RT': row_['RT'],
                'Adduct group': clust_flag
            }, index=[0])

            group = pd.concat([group, line])

        return group

    def _get_unified_struct_groups(self):
        """
        Private method.
        Unifies 'Isotopologue group' and 'Adduct group' into a single 'Structural group'.
        """

        for index, frame in enumerate(self.assay_links):

            frame_ = frame.copy()  # Copy the input pandas.DataFrame

            # Stack Isotopologue and adduct clusters into a single variable 'Structural cluster'
            offset = frame_['Isotopologue group'].max()  # Get the highest number of iso group
            frame_['Adduct group'] = frame_['Adduct group'].apply(lambda x: x+offset)  # adduct clusters above iso
            adduct_frame = frame_.dropna(subset=['Adduct group'])  # Get only rows with non 0 clusters
            iso_frame = frame_.dropna(subset=['Isotopologue group'])  # Get only rows with non 0 clusters
            adduct_frame.insert(0, 'Structural cluster', adduct_frame['Adduct group'])  # Create structural cluster col
            iso_frame.insert(0, 'Structural cluster', iso_frame['Isotopologue group'])  # Create structural cluster col
            adduct_frame = adduct_frame.loc[:, ['Feature', 'Structural cluster']]
            iso_frame = iso_frame.loc[:, ['Feature', 'Structural cluster']]
            stacked_frames = pd.DataFrame(np.vstack([adduct_frame, iso_frame]), columns=iso_frame.columns)
            frame_ = frame_.merge(stacked_frames, on='Feature', how='outer')

            # Delete non-unique values
            non_unique_features = frame_['Feature'][frame_['Feature'].duplicated(keep=False)].unique()
            for item in non_unique_features:
                # For all non-unique features, find all clusters they belong too
                fr_ = frame_[frame_['Feature'] == item].loc[:, ['Feature', 'Structural cluster']]
                # unify all overlapping cluster by assigning the lowest cluster values to all clusters
                for i in range(len(fr_) - 1):
                    # Avoid future warning by opting into the future behavior that disables silent downcasting:
                    if pd.__version__ > '2.2.0': 
                        pd.set_option('future.no_silent_downcasting', True)
                    frame_['Structural cluster'] = frame_['Structural cluster'].replace({fr_.iloc[i + 1, 1]: fr_.iloc[0, 1]})
            unified_frame = frame_.drop_duplicates(subset='Feature')  # Delete non unique Features

            # Update the current assay metadata with the unified structural clusters
            self.assay_links[index] = unified_frame


    def _get_annotation(self, roi=None):
        """
        Private method.
        Annotates hypothetical features using the RIO files from the National Phenome Centre.

        Args:
            roi (pandas.DataFrame, optional): Region of Interest (RIO) file from the National Phenome Centre. Defaults to None.

        Returns:
            pandas.DataFrame: DataFrame with structural clusters and feature annotations.
        """

        # Load RIO files
        for index, frame in enumerate(self.assay_links):

            frame = frame.copy()
                
            if roi is None:
                if frame.loc[0, 'Assay'] == 'HPOS':
                    roi_stream = pkg_resources.resource_stream(__name__, 'Data/ROI/HPOS_ROI_V_3_2_1.csv')
                    roi = pd.read_csv(roi_stream, encoding='windows-1252')

                elif frame.loc[0, 'Assay'] == 'UHPOS':
                    roi_stream = pkg_resources.resource_stream(__name__, 'Data/ROI/HPOS_ROI_V_3_2_1.csv')
                    roi = pd.read_csv(roi_stream, encoding='windows-1252')

                elif frame.loc[0, 'Assay'] == 'SHPOS':
                    roi_stream = pkg_resources.resource_stream(__name__, 'Data/ROI/HPOS_ROI_V_3_2_1.csv')
                    roi = pd.read_csv(roi_stream, encoding='windows-1252')

                elif frame.loc[0, 'Assay'] == 'LPOS':
                    roi_stream = pkg_resources.resource_stream(__name__, 'Data/ROI/LPOS_ROI_V_5_1_2.csv')
                    roi = pd.read_csv(roi_stream, encoding='windows-1252')

                elif frame.loc[0, 'Assay'] == 'LNEG':
                    roi_stream = pkg_resources.resource_stream(__name__, 'Data/ROI/LNEG_ROI_V_5_1_1.csv')
                    roi = pd.read_csv(roi_stream, encoding='windows-1252')

                elif frame.loc[0, 'Assay'] == 'RPOS':
                    roi_stream = pkg_resources.resource_stream(__name__, 'Data/ROI/RPOS_ROI_V_3_2_0.csv')
                    roi = pd.read_csv(roi_stream, encoding='windows-1252')

                elif frame.loc[0, 'Assay'] == 'RNEG':
                    roi_stream = pkg_resources.resource_stream(__name__, 'Data/ROI/RNEG_ROI_V_3_2_0.csv')
                    roi = pd.read_csv(roi_stream, encoding='windows-1252')

                else:
                    raise Exception("ANNOTATION ERROR - Assay has not been recognised. Please ensure that assay name "
                                    "matches naming convention of the National Phenome Centre (NPC) or run structural "
                                    "search without annotation. For more information on NPC assay naming convention please "
                                    "visit https://github.com/phenomecentre/npc-open-lcms")

            else:
                roi = roi

            # Create copies of = DataFrames
            frame_ = frame.copy()
            frame_.reset_index(inplace=True, drop=True)
            roi_ = roi.copy()
            roi_.reset_index(inplace=True, drop=True)
            # Create empty pandas.DataFrame for annotations
            _annotations = pd.DataFrame(columns=['cpdName', 'chemicalFormula', 'ion', 'mzMin', 'mzMax', 'rtMin', 'rtMax'])
            # Iterate through all features
            for i in range(len(frame_)):
                id_current = frame_.loc[i, 'Feature']
                rt_current = frame_.loc[i, 'RT']  # variable used in the query
                mz_current = frame_.loc[i, 'm/z']  # variable used in query
                query = "mzMin < @mz_current < mzMax & rtMin < @rt_current < rtMax"  # query to search the pandas.DataFrame
                retrieved = roi_.query(query, inplace=False)  # query the pandas.DataFrame
                retrieved.reset_index(inplace=True, drop=True)  # reindex allows to use .loc[]
                if retrieved.shape[0] > 0:  # If any annotations found
                    line = pd.DataFrame({
                        'Feature': id_current,
                        'cpdName': retrieved.loc[0, 'cpdName'],
                        'chemicalFormula': retrieved.loc[0, 'chemicalFormula'],
                        'ion': retrieved.loc[0, 'ion'],
                        'mzMin': retrieved.loc[0, 'mzMin'],
                        'mzMax': retrieved.loc[0, 'mzMax'],
                        'rtMin': retrieved.loc[0, 'rtMin'],
                        'rtMax': retrieved.loc[0, 'rtMax']
                    }, index=[0])
                    # add statement to avoid future warnings
                    if len(_annotations) == 0:
                        _annotations = line.copy()
                    else:
                        _annotations = pd.concat([_annotations, line])

            # If any annotation found, then save them merge them with original pandas.DataFrame
            if _annotations.shape[0] > 0:
                frame_annotation = frame_.merge(_annotations, on='Feature', how='left')
            else:
                # If no annotations were found, then add to the frame_ empty columns from _annotations
                frame_annotation = frame_.reindex(frame_.columns.union(_annotations.columns), axis=1)
                frame_annotation = frame_annotation.reindex(frame_.columns.tolist() + _annotations.columns.tolist(), axis=1)

            self.assay_links[index] = frame_annotation
            # return frame_annotation

    def _get_cross_assay_links(self):
        """
        Private method.
        Search for cross-assay links between features based on their m/z and RT. 
        Features that are part of structural cluster are linked only on their M+H and M-H adducts.
        For singletons neutral mass estimated from M+H and M-H is used for cross-assay linking.
        """
        
        # Load data
        struct_data = self.structural_links
        signletons_data = self.structural_links
        data_original = self.structural_links

        # Adduct groups
        # select only the rows where the structural cluster is not NaN
        struct_data = struct_data.dropna(subset=['Structural cluster'])
        # get only rows where adduct is like M+H or M-H
        struct_data = struct_data[struct_data['Adduct'].str.contains(r'\[M\+H\]|\[M-H\]', regex=True, na=False)]
        # delete duplicates of Expected neutral mass
        struct_data = struct_data.drop_duplicates(subset='Expected neutral mass')
        struct_data.reset_index(inplace=True, drop=True)

        # Singletons and Isotopologues
        # calculate expected neutral mass for singletons
        signletons_data = signletons_data.copy()
        # frome singletons_data get only rows where the structural cluster is NaN
        signletons_data = signletons_data[signletons_data['Expected neutral mass'].isna()]
        # For column 'Isotopologue group' keep only rows where the values is 0 or NaN
        signletons_data = signletons_data[((signletons_data['Isotopologue pattern'] == 0) 
                                         | (pd.isna(signletons_data['Isotopologue pattern'])))]
        signletons_data.reset_index(inplace=True, drop=True)
        # if assay is like POS then add 1.007276466812 to the m/z value
        expected = signletons_data.apply(lambda x: x['m/z'] - 1.007276466812 if re.search(r'POS', x['Assay'], re.IGNORECASE) 
                                         else x['m/z'] + 1.007276466812, axis=1)
        expected.reset_index(inplace=True, drop=True)
        signletons_data.loc[:, 'Expected neutral mass'] = expected

        # Merge Adduct and Singletons data
        data = pd.concat([struct_data, signletons_data])
        data.reset_index(inplace=True, drop=True)

        # Create an empty column and isotopologue group flag
        data['Cross-assay link'] = [np.nan] * data.shape[0]
        cross_assay_flag = 1

        # loop through all rows and check if the expected neutral mass for i the same as for j
        for i in range(len(data)):
            for j in range(len(data)):

                # check if the Assay is different
                if data.iloc[i, 1] != data.iloc[j, 1]:
                    PPM_diff = self._mean_ppm_diff(data.loc[i, 'Expected neutral mass'], data.loc[j, 'Expected neutral mass'])
                    RT_diff = abs(data.loc[i, 'RT'] - data.loc[j, 'RT'])

                    # check if the two is LPOS and LNEG or RPOS and RNEG
                    if ((re.search(r'LPOS', data.iloc[i, 1], re.IGNORECASE) and
                        re.search(r'LNEG', data.iloc[j, 1], re.IGNORECASE)) or
                        (re.search(r'RPOS', data.iloc[i, 1], re.IGNORECASE) and 
                         re.search(r'RNEG', data.iloc[j, 1], re.IGNORECASE))
                    ):
                        if PPM_diff <= self.ppm and RT_diff <= self.rt_win:
                            if np.isnan(data.loc[i, 'Cross-assay link']) and np.isnan(data.loc[j, 'Cross-assay link']):
                                data.loc[i, 'Cross-assay link'] = cross_assay_flag
                                data.loc[j, 'Cross-assay link'] = cross_assay_flag
                                cross_assay_flag += 1
                            elif np.isnan(data.loc[i, 'Cross-assay link']):
                                    data.loc[i, 'Cross-assay link'] = data.loc[j, 'Cross-assay link']
                            elif np.isnan(data.loc[j, 'Cross-assay link']):
                                    data.loc[j, 'Cross-assay link'] = data.loc[i, 'Cross-assay link']
                    
                    # If for pair with different LC
                    else:
                        if PPM_diff <= self.ppm:
                            if np.isnan(data.loc[i, 'Cross-assay link']) and np.isnan(data.loc[j, 'Cross-assay link']):
                                data.loc[i, 'Cross-assay link'] = cross_assay_flag
                                data.loc[j, 'Cross-assay link'] = cross_assay_flag
                                cross_assay_flag += 1
                            elif np.isnan(data.loc[i, 'Cross-assay link']):
                                    data.loc[i, 'Cross-assay link'] = data.loc[j, 'Cross-assay link']
                            elif np.isnan(data.loc[j, 'Cross-assay link']):
                                    data.loc[j, 'Cross-assay link'] = data.loc[i, 'Cross-assay link']

        # create new dataframe that only contains 'Feature' and 'Cross-assay link'
        cross_assay_links = data.loc[:, ['Feature', 'Cross-assay link']]

        # add cross-assay links to the original dataframe
        data = data_original.merge(cross_assay_links, on='Feature', how='left')

        self.structural_links = data

    @staticmethod
    def _mean_ppm_diff(x, y):
        """
        Private method.
        Calculates the mean PPM difference between two numbers.

        Args:
            x (float): The first number.
            y (float): The second number.

        Returns:
            float: Mean PPM difference between `x` and `y`.
        """

        diff1 = abs((x - y) / x * 1000000)
        diff2 = abs((y - x) / y * 1000000)
        return (diff1 + diff2) / 2
    
    def get_correlation_clusters(self, flat_method ='constant', cut_threshold=0.7, max_clusters=5, cor_method='pearson' ,
                                 linkage_method='complete', metric='euclidean', **kwargs):
        """
        Clusters features based on their correlations. The method uses hierarchical clustering to create clusters.
        To flatten clusters, the method uses either a constant threshold or silhouette score.

        Args:
            flat_method (str {'constant', 'silhouette'}, optional):
                Method for cluster flattening:
                - 'constant': Flattens clusters based on a constant threshold (cut_threshold).
                - 'silhouette': Flattens clusters based on most optimal silhouette score.
                Defaults to 'constant'.
            cut_threshold (float, optional): Constant threshold for flattening clusters. Defaults to 0.7.
            max_clusters (int, optional): Maximum number of clusters for silhouette method. Defaults to 5.
            cor_method (str {'pearson', 'kendall', 'spearman'}, optional): Method for calculation correlations. Defaults to 'pearson'.
            linkage_method (str, optional): The linkage criterion determines which distance to use between sets of observation.
                The algorithm will merge the pairs of cluster that minimise this criterion.
                - 'single': Single linkage minimises the maximum distance between observations of pairs of clusters.
                - 'complete': Complete linkage minimises the maximum distance between observations of pairs of clusters.
                - 'average': Average linkage minimises the average of the distances between all observations of pairs of clusters.
                - 'ward': Ward minimises the variance of the clusters being merged.
                - 'weighted': Weighted linkage minimises the sum of the product of the distances and the number of observations in pairs of clusters.
                    Only available for 'constant' flatting method.
                - 'centroid': Centroid linkage minimises the distance between the centroids of clusters.
                    Only available for 'constant' flatting method.
                - 'median': Median linkage minimises the distance between the medians of clusters.
                    Only available for 'constant' flatting method.
                Defaults to 'complete'.
            metric (str, optional): The distance metric to use. The metric to use when calculating distance between instances in a feature array.
                Metric used to compute the linkage. Can be “euclidean”, “l1”, “l2”, “manhattan”, “cosine”, or “precomputed”.
                If linkage is “ward”, only “euclidean” is accepted. If “precomputed”, a distance matrix is needed as input for the fit method.
                Defaults to 'euclidean'.
        """
       
        # Check if data were loaded
        if self.structural_links is None:
            warnings.simplefilter('error', RuntimeWarning)
            warnings.warn("You have to first run structural searcher.",
                          RuntimeWarning)

        # Validate the input method
        # List of acceptable arguments
        acceptable_flat_methods = ['constant', 'silhouette']

        if flat_method not in acceptable_flat_methods:
            raise ValueError(f"Invalid flatting method '{flat_method}'. Choose from {acceptable_flat_methods}.")

        # Calculate correlation and similarity
        df = self.intensities
        # Calculate correlation between variables
        correlation = df.corr(method=cor_method)
        # Calculate dissimilarity
        dissimilarity = 1 - abs(correlation)

        # Reorder the structural links by original metadata order
        self.structural_links = self.structural_links.merge(self.feature_metadata[['Feature']], on='Feature', how='right')

        if flat_method == 'constant':
        # Check if metadata have been loaded
            z = linkage(squareform(dissimilarity), method=linkage_method, metric=metric)
            # Get flat clusters
            f_clust = fcluster(z, t=cut_threshold, criterion='distance', depth=100)
            correlation_clusters = pd.DataFrame({ 'Feature': df.columns, 'Correlation cluster':f_clust})

            # Plot correlation heatmap and dendrogram
            plt.subplots(figsize=(17, 10))
            sns.heatmap( data=correlation, annot=False, annot_kws={"fontsize":6})
            plt.show()
            plt.figure(figsize=(17.3, 5))
            dendrogram( Z=z, labels=df.columns, orientation='top', leaf_rotation=90, color_threshold=cut_threshold, **kwargs)
            plt.show()
            
            # Save data
            self.correlation_clusters = correlation_clusters
            self.structural_links['Correlation cluster'] = f_clust

        elif flat_method == 'silhouette':

            # Replace 'max_clusters' with the maximum number of clusters you want to consider
            cluster_range = range(2, max_clusters + 1)
            # Initialize an empty dictionary to store silhouette scores for each number of clusters
            silhouette_scores = {}
            # Iterate over the range of cluster numbers
            for n_clusters in cluster_range:
                # Create an AgglomerativeClustering model with 'n_clusters'
                model = AgglomerativeClustering(n_clusters=n_clusters, metric='precomputed', linkage=linkage_method)

                # Fit the model to the dissimilarity matrix
                cluster_labels = model.fit_predict(dissimilarity)

                # Calculate the silhouette score for the current number of clusters
                silhouette_scores[n_clusters] = silhouette_score(dissimilarity, cluster_labels, metric='precomputed')

            # Find the number of clusters with the highest silhouette score
            best_n_clusters = max(silhouette_scores, key=silhouette_scores.get)
            best_score = silhouette_scores[best_n_clusters]
            # Print the best number of clusters and its silhouette score
            print(f"Best number of clusters based on silhouette score: {best_n_clusters}")
            print(f"Silhouette score for {best_n_clusters} clusters: {best_score}")
            # Now you can use 'best_n_clusters' to create flat clusters
            model = AgglomerativeClustering(n_clusters=best_n_clusters, metric='precomputed', linkage=linkage_method)
            flat_clusters = model.fit_predict(dissimilarity)

            # Save data
            self.correlation_clusters = flat_clusters
            self.structural_links['Correlation cluster'] = flat_clusters


            # Visualisation
            plt.subplots(figsize=(17, 10))
            sns.heatmap( data=correlation, annot=False, annot_kws={"fontsize":6})
            plt.show()

            # Compute the silhouette scores for each sample
            silhouette_vals = silhouette_samples(dissimilarity, flat_clusters, metric='precomputed')
            # Compute the mean silhouette score
            silhouette_avg = silhouette_score(dissimilarity, flat_clusters, metric='precomputed')
            # Plotting the silhouette plot
            fig, ax = plt.subplots()
            y_lower = 10
            for i in range(best_n_clusters):
                # Aggregate the silhouette scores for samples belonging to cluster i
                ith_cluster_silhouette_vals = silhouette_vals[flat_clusters == i]
                ith_cluster_silhouette_vals.sort()

                size_cluster_i = ith_cluster_silhouette_vals.shape[0]
                y_upper = y_lower + size_cluster_i

                color = plt.cm.nipy_spectral(float(i) / best_n_clusters)
                ax.fill_betweenx(np.arange(y_lower, y_upper),
                                0, ith_cluster_silhouette_vals,
                                facecolor=color, edgecolor=color, alpha=0.7)

                # Label the silhouette plots with their cluster numbers at the middle
                ax.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))
                # Compute the new y_lower for next plot
                y_lower = y_upper + 10  # 10 for the 0 samples

            ax.set_title("The silhouette plot for best number of clusters.")
            ax.set_xlabel("The silhouette coefficient values")
            ax.set_ylabel("Cluster label")
            # The vertical line for average silhouette score of all the values
            ax.axvline(x=silhouette_avg, color="red", linestyle="--")
            ax.set_yticks([])  # Clear the yaxis labels / ticks
            ax.set_xticks(np.arange(-0.1, 1.1, 0.2))
            plt.show()


    def get_structural_network(self, include_all=False, interactive=False, 
                               return_nx_object=False,  output_file='interactive.html', labels=False, master_file=None):
        """
        Generates a structural network graph based on the provided master file or the loaded structural links data.
        The method creates a network graph based on the generated structural data or the provided master file.
        The network graph includes nodes representing features and edges representing different types of links.
        The graph can be displayed interactively using pyvis.network or using NetworkX and matplotlib.
        The graph can be saved as a NetworkX object if return_nx_object is True.

        Args:
            include_all (bool, optional): Whether to include all features in the network, even if they are not structurally linked to other features.
                Defaults to False.
            interactive (bool, optional): Whether to display the network graph interactively using pyvis.network.
                If False, the network graph is displayed using NetworkX and Matplotlib.
                Defaults to False.
            return_nx_object (bool, optional): Whether to return the NetworkX object representing the network graph edited in CytoScape. 
                Defaults to False.
            output_file (str, optional): The name of the output file when displaying the network graph interactively using pyvis.network.
                Only applicable when interactive is True. 
                Defaults to 'interactive.html'.
            labels (bool, optional): Whether to display labels for the nodes in the network graph.
                Only applicable when interactive is False. 
                Defaults to False.
            master_file (pd.DataFrame, optional): The master file containing necessary columns for generating the network.
                This is intended for cases when structural links required manual curation (e.g. manually assigned isotopologue groups, adduct groups, etc.)
                If not provided, the function uses the loaded structural links data.
                **Required columns:** 
                - 'Feature': Feature ID (e.g. HPOS_233.25_149.111m/z). Required.
                - 'Assay': Assay name (e.g. HPOS). Required.
                - 'Isotopologue group': Grouped features with similar isotopologue patterns (e.g. 1, 2 ... N). Required.
                - 'Isotopologue pattern': Representing M+0, M+1, M+2 ... M+N for each isotopologue group (e.g. 0, 1, 2 ... N). Required.
                - 'Adduct group': Grouped features with similar adduct patterns (e.g. 1, 2 ... N). Required.
                - 'Adduct': Assigned adduct labels to all features within an adduct group (e.g. [M+H]+, [M-H]-). Required.
                - 'Structural cluster': Grouped features based on the intersection of isotopologue and adduct clusters (e.g. 1, 2, 3 ... N). Required.
                - 'Correlation cluster': Grouped flattened correlation clusters (e.g. 1, 2, 3 ... N). Required.
                - 'Cross-assay link': Linked structural clusters between different assays. Required.
                - 'cpdName': Compound name (e.g. Caffeine). Optional.
                Defaults to None.

        Returns:
            NetworkX.Graph or None: The NetworkX object representing the network graph, if return_nx_object is True.
                Edge weights represent the type of link between features:
                - Isotopologue: 1
                - Adduct: 5
                - Cross-assay link: 10
                Otherwise, None.
                                    

        Raises:
            RuntimeWarning: If no data is loaded and no master file is provided.
            RuntimeWarning: If the provided master file is missing necessary columns.

        """
        
        # Check if data have been loaded
        if master_file is None and self.structural_links is None:
            warnings.simplefilter('error', RuntimeWarning)
            warnings.warn("No data loaded. Use 'get_structural_clusters()' to create structural links file or load a " 
                          "master file using the 'master_file' parameter.",
                          RuntimeWarning)
        
        # Import master if 
        if master_file is not None:
            # Define necessary columns for generating the network
            required_columns = ['Feature',
                                 'Assay',
                                 'Isotopologue group',
                                 'Isotopologue pattern',
                                 'Adduct group',
                                 'Adduct',
                                 'Structural cluster',
                                 'Correlation cluster', 
                                 'Cross-assay link']
            missing_columns = [col for col in required_columns if col not in master_file.columns]
            if not missing_columns:
                master = master_file.copy()
            else:
                warnings.simplefilter('error', RuntimeWarning)
                warnings.warn(f"Missing columns in the master file: {missing_columns}",
                          RuntimeWarning)
        else:
            master = self.structural_links.copy()

        # Drop all features that are not structurally linked to any other feature
        if not include_all:
            master.dropna(subset=['Structural cluster', 'Cross-assay link'], how='all', inplace=True)

       # Optional Columns for the network
        if 'cpdName' not in master.columns:
            master['cpdName'] = np.nan 

        master['Node'] = master['Feature']  # Create a node column

        # Create a network graph
        G = nx.Graph()
        for index, row in master.iterrows():
            G.add_node(row['Node'], 
                       Assay=row['Assay'], 
                       Isotopologue_group=row['Isotopologue group'],
                       Isotopologue_pattern=row['Isotopologue pattern'], 
                       Adduct_group=row['Adduct group'],
                       Adduct=row['Adduct'], 
                       Structural_cluster=row['Structural cluster'],
                       Correlation_cluster=row['Correlation cluster'], 
                       Cross_assay_link=row['Cross-assay link'],
                       Annotation=row['cpdName']
                       )

        # Add edges to the graph
        for u in G.nodes():
            for v in G.nodes():
                if (
                    u != v and G.nodes[u]['Structural_cluster'] == G.nodes[v]['Structural_cluster'] and 
                    ( G.nodes[u]['Adduct_group'] == G.nodes[v]['Adduct_group'] )
                ):
                    G.add_edge(u, v, weight = 5, type= 5)  # Adduct weight = 5

                if u != v and G.nodes[u]['Cross_assay_link'] == G.nodes[v]['Cross_assay_link']:
                    G.add_edge(u, v, weight = 10, type= 10)  # Cross assay link weight = 10

                if (u != v and 
                    G.nodes[u]['Structural_cluster'] == G.nodes[v]['Structural_cluster'] and 
                    (G.nodes[u]['Isotopologue_group'] == G.nodes[v]['Isotopologue_group'] and 
                    G.nodes[u]['Isotopologue_pattern'] == G.nodes[v]['Isotopologue_pattern']+1 )
                ):
                    G.add_edge(u, v, weight = 1, type= 1)  # Isotopologue weight = 1

        # Define colours
        correlation_clusters = set(nx.get_node_attributes(G, 'Correlation_cluster').values())
        colour_map = {cluster: plt.cm.tab10(i) for i, cluster in enumerate(correlation_clusters)}

        # Chose plot option between NetworkX and pyvis.network 
        if  interactive:
            iG = G.copy()
            # create vis network
            net = Network(notebook=True, cdn_resources='remote')
            # load the NetworkX graph
            net.from_nx(iG)
            net.show(output_file)
            display(IFrame(output_file, width="100%", height="600px"))

        else:
            pos = nx.spring_layout(G,  threshold=0.015)
            node_colors = [colour_map[G.nodes[node]['Correlation_cluster']] for node in G.nodes()]
            nx.draw(G, pos, with_labels=labels, font_size=6, node_color=node_colors, node_size=300, edge_color='gray')
            plt.title("Structural Network")
            plt.show()
        
        # Save NetworkX object
        if return_nx_object:
            return G



# -*- coding: utf-8 -*-
#
# Authors: Lukas Kopecky <l.kopecky22@imperial.ac.uk>
#          Timothy MD Ebbels 
#          Elizabeth J Want
#
# License: BSD 3-clause

import pytest
import numpy as np
import pandas as pd
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from mamsi.mamsi_pls import MamsiPls


# Fixtures
## To work with single block data
@pytest.fixture
def sample_data():
    np.random.seed(42)
    x = np.random.rand(100, 10)
    y = np.random.randint(0, 2, 100)
    return x, y

## To work with multi-block data
@pytest.fixture
def sample_multiblock_data():
    np.random.seed(42)
    x1 = np.random.rand(100, 10)
    x2 = np.random.rand(100, 15)
    y = np.random.randint(0, 2, 100)
    return [x1, x2], y

## To work with multi-block data in pandas DataFrame format
@pytest.fixture
def sample_multiblock_data_df():
    np.random.seed(42)
    x1 = pd.DataFrame(np.random.rand(100, 10))
    x2 = pd.DataFrame(np.random.rand(100, 15))
    y = np.random.randint(0, 2, 100)
    return [x1, x2], y

# Tests
## Initialisation
def test_initialisation():
    model = MamsiPls(n_components=3)
    assert model.n_components == 3

# Test inherited methods from the mbpls parent class
@pytest.mark.parametrize("data_fixture", ["sample_data", "sample_multiblock_data", "sample_multiblock_data_df"])
def test_fit(request, data_fixture):
    x, y = request.getfixturevalue(data_fixture)
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    assert hasattr(model, 'beta_')

@pytest.mark.parametrize("data_fixture", ["sample_data", "sample_multiblock_data", "sample_multiblock_data_df"])
def test_predict(request, data_fixture):
    x, y = request.getfixturevalue(data_fixture)
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    y_pred = model.predict(x)
    assert len(y_pred) == len(y)

# Test MamsiPls specific methods
## test evaluate_class_model
@pytest.mark.parametrize("data_fixture", ["sample_data", "sample_multiblock_data", "sample_multiblock_data_df"])
def test_evaluate_class_model(request, data_fixture):
    x, y = request.getfixturevalue(data_fixture)
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    y_pred = model.evaluate_class_model(x, y)
    assert len(y_pred) == len(y)

## test evaluate_regression_model
@pytest.mark.parametrize("data_fixture", ["sample_data", "sample_multiblock_data", "sample_multiblock_data_df"])
def test_evaluate_regression_model(request, data_fixture):
    x, y = request.getfixturevalue(data_fixture)
    y = np.random.rand(100)  # Continuous target for regression
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    y_pred = model.evaluate_regression_model(x, y)
    assert len(y_pred) == len(y)

## test kfold_cv
@pytest.mark.parametrize("data_fixture", ["sample_data", "sample_multiblock_data", "sample_multiblock_data_df"])
def test_kfold_cv(request, data_fixture):
    x, y = request.getfixturevalue(data_fixture)
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    scores = model.kfold_cv(x, y, n_splits=3, n_jobs=1)
    assert not scores.empty

## test montecarlo_cv
@pytest.mark.parametrize("data_fixture", ["sample_data", "sample_multiblock_data", "sample_multiblock_data_df"])
def test_montecarlo_cv(request, data_fixture):
    x, y = request.getfixturevalue(data_fixture)
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    scores = model.montecarlo_cv(x, y, repeats=5, n_jobs=1)
    assert not scores.empty

# test montecarlo_cv with different number of repeats
@pytest.mark.parametrize("repeats", [1, 5, 10])
def test_montecarlo_cv_repeats(sample_multiblock_data, repeats):
    x, y = sample_multiblock_data
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    scores = model.montecarlo_cv(x, y, repeats=repeats, n_jobs=1)
    assert not scores.empty

# test montecarlo_cv with different data sizes
@pytest.mark.parametrize("data_size", [(50, 5), (200, 20)])
def test_montecarlo_cv_data_sizes(data_size):
    np.random.seed(42)
    x1 = np.random.rand(data_size[0], data_size[1])
    x2 = np.random.rand(data_size[0], data_size[1] + 5)
    y = np.random.randint(0, 2, data_size[0])
    x = [x1, x2]
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    scores = model.montecarlo_cv(x, y, repeats=5, n_jobs=1)
    assert not scores.empty

# test montecarlo_cv with empty data blocks
def test_montecarlo_cv_empty_data():
    x = [np.empty((0, 10)), np.empty((0, 15))]
    y = np.empty((0,))
    model = MamsiPls(n_components=2)
    with pytest.raises(ValueError):
        model.fit(x, y)
        model.montecarlo_cv(x, y, repeats=5, n_jobs=1)

# test montecarlo_cv with different random states
@pytest.mark.parametrize("random_state", [0, 42, 100])
def test_montecarlo_cv_random_state(sample_multiblock_data, random_state):
    x, y = sample_multiblock_data
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    scores_1 = model.montecarlo_cv(x, y, repeats=5, random_state=random_state, n_jobs=1)
    scores_2 = model.montecarlo_cv(x, y, repeats=5, random_state=random_state, n_jobs=1)
    assert scores_1.equals(scores_2)

# test mb_vip_permtest with different number of permutations
@pytest.mark.parametrize("n_permutations", [10, 50, 100])
def test_mb_vip_permtest_permutations(sample_multiblock_data, n_permutations):
    x, y = sample_multiblock_data
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    p_vals = model.mb_vip_permtest(x, y, n_permutations=n_permutations, n_jobs=1)
    assert p_vals is not None
    assert len(p_vals) == sum([block.shape[1] for block in x])

# test mb_vip_permtest with return_scores=True
def test_mb_vip_permtest_return_scores(sample_multiblock_data):
    x, y = sample_multiblock_data
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    p_vals, vip_scores = model.mb_vip_permtest(x, y, n_permutations=100, return_scores=True, n_jobs=1)
    assert p_vals is not None
    assert vip_scores is not None
    assert len(p_vals) == sum([block.shape[1] for block in x])
    assert vip_scores.shape[1] == 100

# test mb_vip_permtest with different data formats
@pytest.mark.parametrize("data_fixture", ["sample_multiblock_data", "sample_multiblock_data_df"])
def test_mb_vip_permtest_data_formats(request, data_fixture):
    x, y = request.getfixturevalue(data_fixture)
    model = MamsiPls(n_components=2)
    model.fit(x, y)
    p_vals = model.mb_vip_permtest(x, y, n_permutations=10, n_jobs=1)
    assert p_vals is not None
    assert len(p_vals) == sum([block.shape[1] for block in x])

# test mb_vip_permtest with different n_components
@pytest.mark.parametrize("n_components", [1, 2, 3])
def test_mb_vip_permtest_n_components(sample_multiblock_data, n_components):
    x, y = sample_multiblock_data
    model = MamsiPls(n_components=n_components)
    model.fit(x, y)
    p_vals = model.mb_vip_permtest(x, y, n_permutations=10, n_jobs=1)
    assert p_vals is not None
    assert len(p_vals) == sum([block.shape[1] for block in x])


# -*- coding: utf-8 -*-
#
# Authors: Lukas Kopecky <l.kopecky22@imperial.ac.uk>
#          Timothy MD Ebbels 
#          Elizabeth J Want
#
# License: BSD 3-clause

import pytest
import numpy as np
import pandas as pd
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from mamsi.mamsi_struct_search import MamsiStructSearch

@pytest.fixture
def sample_lcms_data():
    data = {
        'HPOS_233.25_149.111m/z': [100, 200, 300],
        'HPOS_233.25_150.111m/z': [150, 250, 350],
        'HPOS_234.25_149.111m/z': [200, 300, 400]
    }
    return pd.DataFrame(data)

@pytest.fixture
def sample_msi_data():
    data = {
        'Assay1_149.111': [100, 200, 300],
        'Assay1_150.111': [150, 250, 350],
        'Assay2_149.111': [200, 300, 400]
    }
    return pd.DataFrame(data)

def test_load_lcms(sample_lcms_data):
    searcher = MamsiStructSearch()
    searcher.load_lcms(sample_lcms_data)
    assert searcher.intensities.equals(sample_lcms_data)
    assert searcher.feature_metadata is not None
    assert searcher.assay_links is not None

def test_load_msi(sample_msi_data):
    searcher = MamsiStructSearch()
    searcher.load_msi(sample_msi_data)
    assert searcher.intensities.equals(sample_msi_data)
    assert searcher.feature_metadata is not None
    assert searcher.assay_links is not None





from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="mamsi",
    version="1.0.4",
    packages=find_packages(),
    url="https://github.com/kopeckylukas/py-mamsi",
    license="BSD 3-Clause License",
    license_files=["LICENSE.txt"],
    package_data={'': ['Data/Adducts/*', 'Data/ROI/*']},
    author="Lukas Kopecky",
    author_email="l.kopecky22@imperial.ac.uk",
    long_description=long_description,
    long_description_content_type='text/markdown',
    description="Multi-Assay Mass Spectrometry Integration (MAMSI) for multimodal LC-MS data.",
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],

    install_requires=[
        'mbpls==1.0.4',
        'pandas',
        'numpy',
        'matplotlib',
        'scipy',
        'scikit-learn<=1.5.2',
        'seaborn',
        'networkx',
        'pyvis',
        'joblib',
    ],

)